package org.jvnet.updatecenter2;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.MavenProjectHelper;
import org.apache.tools.ant.taskdefs.Expand;
import org.apache.tools.ant.taskdefs.Execute;
import org.apache.tools.ant.Project;

import java.io.File;
import java.io.ByteArrayOutputStream;
import java.io.UnsupportedEncodingException;
import java.io.IOException;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;

/**
 * Builds an IPS package.
 *
 * @goal package
 * @phase package
 * @author Kohsuke Kawaguchi
 */
public class PackageMojo extends AbstractMojo {
    /**
     * Package definition file that drives the packaging process.
     *
     * @parameter default-value="src/proto.py"
     * @required
     */
    protected File proto;

    /**
     * Directory that extracts makepkgs.py to.
     *
     * If this directory already exists, then the script is assumed to be already in there.
     * This is useful for testing modified versions of <tt>makepkgs.py</tt>. This also
     * implements the caching semantics.
     *
     * @parameter default-value="target/makepkgs"
     */
    protected File makePkgsDir;

    /**
     * Local directory to deposit the generated package.
     *
     * @parameter default-value="target/ips-repository"
     */
    protected File localRepository;

    /**
     * The base directory to be passed to <tt>makepkgs.py</tt>.
     * This determines where the relative paths in your <tt>proto.py</tt> are resolved.
     *
     * @parameter default-value="${project.basedir}"
     */
    protected File baseDirectory;

    /**
     * Path to generated IPS package file.
     *
     * @parameter default-value="target/${project.artifactId}-${project.version}.ipstgz"
     */
    protected File packageFile;

    /**
     * Remote repository URL to deposit the generated package.
     *
     * <p>
     * If this parameter is non-null, this takes precedence over
     * the local repository parameter.
     *
     * @parameter
     */
    protected String remoteRepository;

    /**
     * Attaches the created package file as an artifact of the build
     * (which in turn causes it to be installed to your local repository
     * or to be deployed remotely to a Maven repository.)
     *
     * @parameter default-value="true"
     */
    protected boolean attach;

    /**
     * Path to python executable. By default, the mojo invokes "python",
     * which means python will be assumed to be in PATH, but if not,
     * you can override it here.
     *
     * <p>
     * The typical use case of this is to specify it in the command-line,
     * like <tt>mvn -Dpython=/path/to/python install</tt>
     *
     * @parameter expression="${python}" default-value="python"
     */
    protected String python;

    /**
     * This is injected by Maven and not meant to be set by POM.
     *
     * @parameter expression="${project}"
     * @required
     */
    protected MavenProject project;

    /**
     * @component
     */
    private MavenProjectHelper projectHelper;

    public void execute() throws MojoExecutionException, MojoFailureException {
        try {
            File jar = locateMakepkgs();
            unjar(jar,makePkgsDir);
            executeMakePkgs();
        } catch (IOException e) {
            throw new MojoExecutionException("makepkgs failed",e);
        }
    }

    private void executeMakePkgs() throws MojoExecutionException, IOException {
        boolean createdTgz = false;

        List<String> commandLine = new ArrayList<String>();
        commandLine.add(python);
        commandLine.add(new File(makePkgsDir,"makepkgs.py").getAbsolutePath());
        if(remoteRepository!=null) {
            commandLine.add("-s");
            commandLine.add(remoteRepository);
        } else {
            commandLine.add("-d");
            commandLine.add(localRepository.getAbsolutePath());
            commandLine.add("-t");
            commandLine.add(packageFile.getAbsolutePath());
            createdTgz = true;
        }
        commandLine.add("-b");
        commandLine.add(baseDirectory.getAbsolutePath());
        commandLine.addAll(Arrays.asList(new String[]{
            "-D", "groupId="+project.getGroupId(),
            "-D", "artifactId="+project.getArtifactId(),
            "-D", "version="+project.getVersion(),
        }));

        commandLine.add(proto.getAbsolutePath());

        getLog().debug("Invoking "+commandLine);
        Execute exec = new Execute();
        exec.setCommandline(commandLine.toArray(new String[commandLine.size()]));
        exec.setWorkingDirectory(project.getBasedir());
        if(exec.execute()!=0)
            throw new MojoExecutionException("makepkgs failed");

        if(createdTgz && attach)
            projectHelper.attachArtifact(project,"ipstgz",null,packageFile);
    }

    /**
     * Unjars the jar to the specified target directory.
     */
    private void unjar(File jar, File destDir) {
        if(destDir.exists())
            return; // noop

        getLog().info("Extracting IPS to "+destDir);
        destDir.mkdirs();

        Expand expand = new Expand();
        expand.setProject(new Project());
        expand.setSrc(jar);
        expand.setDest(destDir);
        expand.execute();
    }

    /**
     * Locates the jar file that makepkgs scripts are in.
     */
    private File locateMakepkgs() throws MalformedURLException {
        String resURL = getClass().getClassLoader().getResource("makepkgs.py").toString();
        resURL = resURL.substring(4, resURL.lastIndexOf('!')); // cut off jar: and the file name portion
        return new File(decode(new URL(resURL).getPath()));
    }

    /**
     * Decode '%HH'.
     */
    private static String decode(String s) {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        for( int i=0; i<s.length();i++ ) {
            char ch = s.charAt(i);
            if(ch=='%') {
                baos.write(hexToInt(s.charAt(i+1))*16 + hexToInt(s.charAt(i+2)));
                i+=2;
                continue;
            }
            baos.write(ch);
        }
        try {
            return new String(baos.toByteArray(),"UTF-8");
        } catch (UnsupportedEncodingException e) {
            throw new Error(e); // impossible
        }
    }

    private static int hexToInt(int ch) {
        return Character.getNumericValue(ch);
    }
}
